import time
import numpy as np
from numpy.testing import assert_approx_equal
import torch
from tiny_vectordb import VectorCollection

LEN = 512
torch_device = "cuda" if torch.cuda.is_available() else "cpu"
# torch_device = "mps" 

def cos_sim_np(tgt, query):
    """
    target: (N, LEN)
    query: (LEN, )
    """
    assert isinstance(tgt, np.ndarray)
    assert isinstance(query, np.ndarray)
    return np.dot(tgt, query) / (np.linalg.norm(tgt, axis=1) * np.linalg.norm(query))

def search_np(ids: list[str], tgt, query, k):
    assert isinstance(tgt, np.ndarray)
    assert isinstance(query, np.ndarray)
    assert isinstance(k, int)

    scores = cos_sim_np(tgt, query)
    topk_indices = np.argsort(scores)[::-1][:k]
    # return [ids[i] for i in topk_indices], scores[topk_indices]
    return ids[topk_indices], scores[topk_indices]

def cos_sim_torch(tgt, query):
    """
    target: (N, LEN)
    query: (LEN, )
    """
    assert isinstance(tgt, torch.Tensor)
    assert isinstance(query, torch.Tensor)
    return torch.matmul(tgt, query) / (torch.norm(tgt, dim=1) * torch.norm(query))

def search_torch(ids: list[str], tgt, query, k):
    assert isinstance(tgt, torch.Tensor)
    assert isinstance(query, torch.Tensor)
    assert isinstance(k, int)

    scores = cos_sim_torch(tgt, query)
    topk_indices = torch.argsort(scores, descending=True)[:k]
    # return [ids[i] for i in topk_indices], scores[topk_indices]
    return ids[topk_indices.cpu().numpy()], scores[topk_indices]

def compare(n: int, k = 16):
    collection = VectorCollection[float](None, "test", LEN, quite_loading=True)
    print(f"Compare for {n} vectors of length {LEN} numpy | torch-{torch_device}:")

    np.random.seed(0)
    vectors = np.random.rand(n, LEN).tolist()
    ids = [str(x) for x in range(n)]

    query = [float(x) for x in range(LEN)]
    target = vectors

    collection.addBlock(ids, vectors)
    _t = time.time()
    sc_0 = collection._impl.score(query)
    t_0 = time.time() - _t
    print("[Eigen] Matrix multiplication time taken:", t_0, "seconds")
    _t = time.time()
    s0 = collection.search(query, k)
    t_01 = time.time() - _t
    print("[Eigen] Search time taken:", t_01, "seconds")

    target_np = np.array(target)
    query_np = np.array(query)
    _t = time.time()
    sc_1 = cos_sim_np(target_np, query_np)
    t_1 = time.time() - _t
    print("[Numpy] Matrix multiplication time taken:", t_1, "seconds")
    _t = time.time()
    s1 = search_np(np.array(ids), target_np, query_np, k)   # type: ignore
    t_11 = time.time() - _t
    print("[Numpy] Search time taken:", t_11, "seconds")

    target_torch = torch.tensor(target).to(torch_device)
    query_torch = torch.tensor(query).to(torch_device)
    _t = time.time()
    sc_2 = cos_sim_torch(target_torch, query_torch)
    t_2 = time.time() - _t
    print("[PyTorch] Matrix multiplication time taken:", t_2, "seconds")
    _t = time.time()
    s2 = search_torch(np.array(ids), target_torch, query_torch, k)      # type: ignore
    t_21 = time.time() - _t
    print("[PyTorch] Search time taken:", t_21, "seconds")

    step = 1 if n < 1000 else n//100
    for i in range(0, n, step):
        assert_approx_equal(sc_0[i], sc_1[i], significant=4)
    
    for i in range(k):
        assert s0[0][i] == s1[0][i]
        assert_approx_equal(s0[1][i], s1[1][i], significant=4)
        assert s0[0][i] == s2[0][i]
        assert_approx_equal(s0[1][i], s2[1][i], significant=4)

    print("----------------------------------")
    print(f"Matrix mult. superiority: {t_1 / t_0 : .2f} | {t_2 / t_0} times faster", end="\n")
    print(f"Search superiority: {t_11 / t_01 : .2f} | {t_21 / t_01} times faster", end="\n\n")

for i in range(2,6)[::-1]:
    compare(10**i)



from tiny_vectordb import VectorCollection

N = 16
collection = VectorCollection(None, "test", N)

vec = [ float(i) for i in range(N)]
# print("Encoding test:")
enc = collection.encoding.encode(vec)
# print(enc)
dec = collection.encoding.decode(enc)
# print(dec)
assert dec == vec


import pkg_resources, os, shutil
VERSION = pkg_resources.get_distribution("tiny_vectordb").version

__this_dir = os.path.dirname(os.path.abspath(__file__))
__this_dir = os.path.abspath(os.path.realpath(__this_dir))
SRC_DIR = os.path.join(__this_dir, "src")
HEADER_DIR = os.path.join(__this_dir, "include")
CACHE_DIR = os.path.join(__this_dir, "_cache")
if not os.path.exists(CACHE_DIR):
    os.mkdir(CACHE_DIR)

def cleanup():
    """
    Should be called before using `pip uninstall`
    """
    global CACHE_DIR
    if os.path.exists(CACHE_DIR):
        shutil.rmtree(CACHE_DIR)

BUILD_DIR = os.path.join(CACHE_DIR, f"build{VERSION}")
BIN_DIR = os.path.join(BUILD_DIR, "bin")
for _d in [BUILD_DIR, BIN_DIR]:
    if not os.path.exists(_d):
        os.makedirs(_d, exist_ok=True)



from tiny_vectordb import VectorDatabase
import numpy as np
import os

n, LEN = 5, 6

test_db_path = os.path.join(os.path.dirname(__file__), "test.db")
if os.path.exists(test_db_path):
    os.remove(test_db_path)

database = VectorDatabase(test_db_path, [{ "name": "Test", "dimension": LEN, }, { "name": "Hello", "dimension": LEN }])

collection = database.getCollection("Test")

np.random.seed(0)
vectors = np.random.rand(n, LEN).tolist()
ids = [str(x) for x in range(n)]

collection.addBlock(ids, vectors)
print(collection["1"])
collection.deleteBlock(["1", "2"])
collection._impl.update("3", np.zeros(LEN).tolist())

collection._impl.print()

database.commit()


from __future__ import annotations
from typing import Union, TypeVar, Generic, Optional, TypedDict, Optional
from .jit import compile
from .config import BIN_DIR
import sys
from importlib import import_module
from .diskio import SqliteIO

Number = Union[int, float]
NumVar = TypeVar('NumVar', int, float)


class CollectionConfig(TypedDict):
    name: str
    dimension: int

class CompileConfig(TypedDict):
    cxx: str
    additional_compile_flags: list[str]
    additional_link_flags: list[str]

class VectorDatabase(dict[str, "VectorCollection[float]"]):
    VERBOSE: bool = False
    def __init__(self, path: str, collection_configs: list[CollectionConfig], compile_config: Optional[CompileConfig] = None):
        super().__init__()
        self.__database_path = path
        self.__disk_io = SqliteIO(path)
        for config in collection_configs:
            self[config['name']] = VectorCollection[float](
                self, 
                quite_loading=not self.VERBOSE, 
                compile_config=compile_config, 
                **config
                )
        self.__initCollections()
    
    def __initCollections(self):
        """Load collection from disk if exists"""
        table_names = self.disk_io.getTableNames()
        assert set(table_names).issubset(set(self.keys())), \
        f"Database corrupted, existing tables: {table_names}, but only specified collections of: {self.keys()}"
        for name in self.keys():
            if table_names.__contains__(name):
                self.getCollection(name).loadFromDisk()
            self.disk_io.touchTable(name)

    @property
    def disk_io(self):
        return self.__disk_io
    
    @property
    def database_path(self):
        return self.__database_path
    
    def getCollection(self, name: str) -> VectorCollection:
        return self[name]
    
    def getCollectionNames(self) -> list[str]:
        return list(self.keys())
    
    def createCollection(self, collection_config: CollectionConfig) -> VectorCollection:
        """ Insert a new collection to database and create a new table in sqlite database """
        _name = collection_config["name"]
        assert not self.keys().__contains__(_name), f"Collection '{_name}' already exists"
        self[_name] = VectorCollection(self, quite_loading=not self.VERBOSE, **collection_config)
        self.disk_io.touchTable(_name)
        return self[_name]
    
    def deleteCollection(self, name: str):
        assert self.keys().__contains__(name), f"Collection '{name}' not exists"
        del self[name]
        self.disk_io.deleteTable(name)

    def commit(self):
        """
        Commit all changes to sqlite database and commit
        """
        for collection in self.values():
            collection._flush()
        self.disk_io.commit()


class _VectorCollectionEncoding(Generic[NumVar]):
    def encode(self, vector: list[NumVar]) -> str: ...
    def decode(self, enc_vector: str) -> list[NumVar]: ...

class VectorCollection(Generic[NumVar]):
    COMPILE_CONFIG: CompileConfig = {
        "cxx": "g++",
        "additional_compile_flags": [],     # "-march=native", "-mtune=native",
        "additional_link_flags": []
    }

    def __init__(
            self, parent: Optional[VectorDatabase], 
            name: str, 
            dimension: int, 
            quite_loading = True, 
            compile_config: Optional[CompileConfig] = None
            ):
        """
        set parent to None if you don't want to save changes to disk
        """
        if not sys.path.__contains__(BIN_DIR):
            if not quite_loading:
                print("Adding", BIN_DIR, "to sys.path")
            sys.path.append(BIN_DIR)
        if compile_config is None:
            compile_config = self.__class__.COMPILE_CONFIG

        _m_name = compile(dimension, quite = quite_loading, **compile_config)
        self.__name = name
        self.__database = parent
        self.__clib = import_module(_m_name)
        self.__impl = self.__clib.VectorCollectionImpl()
    
    @property
    def name(self):
        return self.__name
    @property
    def database(self):
        return self.__database
    @property
    def clib(self):
        return self.__clib
    @property
    def _impl(self):
        return self.__impl
    @property
    def encoding(self) -> _VectorCollectionEncoding[NumVar]:
        return self.__clib.enc
    
    def addBlock(self, ids: list[str], vectors: list[list[NumVar]]):
        """
        Add a bulk of elements, will raise error if id exists
        """
        self._impl.addBulk(ids, vectors)
    
    def deleteBlock(self, ids: list[str]) -> None:
        """
        Delete a bulk of elements, will raise error if id not exists
        """
        self._impl.deleteBulk(ids)
    
    def setBlock(self, ids: list[str], vectors: list[list[NumVar]]) -> None:
        """
        For every element in ids, set the corresponding vector
        Add if not exists, update if exists
        """
        self._impl.setBulk(ids, vectors)

    def update(self, id: str, vector: list[NumVar]) -> bool:
        """
        Change a vector, will raise error if id not exists
        """
        return self._impl.set(id, vector)

    def has(self, id: str) -> bool:
        """
        Check if id exists
        """
        return self._impl.has(id)
    
    def keys(self) -> list[str]:
        """
        Return all ids in this collection
        """
        return self._impl.getAllIds()

    def get(self, id: str) -> list[NumVar]:
        """
        Get a vector by id, return an empty list if not exists
        """
        return self._impl.get(id)
    
    def getBlock(self, ids: list[str]) -> list[list[NumVar]]:
        """
        Get a bulk of vectors by ids, return list element can be empty if not exists
        """
        return self._impl.getBulk(ids)
    
    def search(self, query: list[NumVar], k: int = -1) -> tuple[list[str], list[float]]:
        """Return a tuple of (ids, scores)"""
        return self._impl.search(query, k)
    
    def loadFromDisk(self) -> None:
        """
        Load all data of the collection from disk, 
        Should be called when the collection is attached to a database and is empty
        """
        if not self.database:
            return 
        if len(self) != 0:
            raise RuntimeError("Collection is not empty, cannot load from disk")
        
        ids, enc_vectors = self.database.disk_io.getTableData(self.name)
        self._impl.addRawEncBulk(ids, enc_vectors)

    def _flush(self) -> bool:
        """
        Load all changes to sqlite database memory, but not save to disk,
        If the collection is not attached to a database, return False
        """
        changes = self._impl.flush()
        if not self.database:
            return False
        ids_add, enc_vectors_add = changes["ADD"]
        for identifier, enc_vector in zip(ids_add, enc_vectors_add):
            self.database.disk_io.insetToTable(self.name, identifier, enc_vector)
        ids_update, enc_vectors_update = changes["UPDATE"]
        for identifier, enc_vector in zip(ids_update, enc_vectors_update):
            self.database.disk_io.updateTable(self.name, identifier, enc_vector)
        for identifier in changes["DELETE"][0]:
            self.database.disk_io.deleteFromTable(self.name, identifier)
        return True

    def __len__(self) -> int:
        return self._impl.size()
    
    def __getitem__(self, id: str) -> Optional[list[NumVar]]:
        return self.get(id)
    


"""
vecotr related modules.
vector database and related functions.
"""
from .wrap import VectorDatabase, VectorCollection, CompileConfig, CollectionConfig
from .config import cleanup
from .jit_utils import autoCompileConfig
from . import jit

__all__ = [
    "VectorDatabase",
    "VectorCollection",
    "CompileConfig",
    "CollectionConfig",
    "autoCompileConfig",
    "cleanup", 
    "jit", 
]


import os
import numpy as np
from tiny_vectordb import VectorDatabase, autoCompileConfig

LEN = 16
N = 16

VectorDatabase.VERBOSE = True
if os.path.exists("test.db"):
    raise FileExistsError("test.db already exists")

print(autoCompileConfig())
database = VectorDatabase(
    "test.db", 
    [{ "name": "test", "dimension": LEN, }],
    compile_config=autoCompileConfig()
    )

collection = database.getCollection("test")
if not collection.has("-1"):
    collection.addBlock(["-1"], [[float(x) for x in range(LEN)]])

np.random.seed(0)
vectors = np.random.rand(N, LEN).tolist()
ids = [str(x) for x in range(N)]

collection.addBlock(ids, vectors)

query = [float(x) for x in range(LEN)]
target = vectors
assert collection.has("-1")
assert collection.get("-1") == [float(x) for x in range(LEN)]

assert not collection.get("100")

collection.deleteBlock(["-1", "0"])
assert not collection.has("-1")
assert not collection.has("0")
assert collection.has("1")

database.createCollection({"name": "hello", "dimension": 128})
assert set(database.keys()) == set(["test", "hello"])
assert set(database.disk_io.getTableNames()) == set(["test", "hello"])

database.deleteCollection("test")
assert list(database.keys()) == ["hello"]
assert database.disk_io.getTableNames() == ["hello"]
database.disk_io.conn.close()
os.remove("test.db")



# https://ninja-build.org/manual.html
from ninja import ninja_syntax
import pybind11
import os, sysconfig, subprocess, platform, sys
from .config import CACHE_DIR, SRC_DIR, HEADER_DIR, BUILD_DIR, BIN_DIR
from .jit_utils import initEigenSrc, checkCommandExists

eigen_version = "3.4.0"
eigen_src_path = os.path.join(CACHE_DIR, f"eigen{eigen_version}")
initEigenSrc(eigen_src_path, eigen_version)

def _writeNinja(
        feat_dim: int, 
        cxx = "g++", 
        additional_compile_flags = [],
        additional_link_flags = []
        ):
    if not checkCommandExists(cxx):
        raise RuntimeError(f"{cxx} not found.")

    module_name = _get_module_name(feat_dim)
    bin_dir = os.path.join(BIN_DIR, module_name)
    script_dir = os.path.join(BUILD_DIR, "scripts_"+ module_name)
    lib_dir = os.path.join(BIN_DIR, "lib")
    ninja_build_file = os.path.join(script_dir, "build.ninja")
    for _d in [bin_dir, lib_dir, script_dir]:
        if not os.path.exists(_d):
            os.makedirs(_d, exist_ok=True)

    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
    obj_suffix = ".obj" if platform.system() == "Windows" else ".o"
    py_includes = sysconfig.get_config_var('INCLUDEPY')

    with open(ninja_build_file, "w") as build_file:
        writer = ninja_syntax.Writer(build_file)
        writer.comment("This file is generated by build.py")

        to_compile = [ "vecdbImpl", ]
        to_compile_lib = [ "diskIO", "b64enc", "searchAlgorithm" ]

        if cxx == "g++" or cxx == "clang++":
            cxx_flags = [
                "-std=c++17",
                "-Wall",
                f"-I{eigen_src_path}",
                f"-I{py_includes}",
                f"-I{pybind11.get_include()}",
                f"-I{HEADER_DIR}",
                f"-DFEAT_DIM={feat_dim}",
                f"-DMODULE_NAME={module_name}",

                ## Optimization flags
                "-DNDEBUG",
                "-O2",
                "-funroll-loops",
            ] + additional_compile_flags

            if platform.system() == "Windows":
                cxx_flags += [ "-DYNAMICBASE" ]
            else:
                cxx_flags += [ "-fPIC" ]

            link_flags = [
                "-shared",
                "-lstdc++",
            ] + additional_link_flags
            if platform.system() == "Darwin":
                link_flags.append("-undefined dynamic_lookup")

            writer.variable("CXX", cxx)
            writer.variable("CXX_FLAGS", " ".join(cxx_flags))
            writer.variable("LINK_FLAGS", " ".join(link_flags))

            writer.rule("compile", "$CXX -MMD -MF $out.d $CXX_FLAGS $in -c -o $out", depfile="$out.d", description="compile $out")
            for _m in to_compile:
                writer.build(os.path.join(bin_dir, f"{_m}{feat_dim}{obj_suffix}"), "compile", os.path.join(SRC_DIR, f"{_m}.cpp"))
            for _m in to_compile_lib:
                writer.build(os.path.join(lib_dir, f"{_m}{obj_suffix}"), "compile", os.path.join(SRC_DIR, f"{_m}.cpp"))
            
            writer.rule("link", "$CXX $LINK_FLAGS $in -o $out", description="link $out")
            writer.build(os.path.join(bin_dir, f"{module_name}{ext_suffix}"), "link", \
                            [os.path.join(bin_dir, f"{_m}{feat_dim}{obj_suffix}") for _m in to_compile] + \
                            [os.path.join(lib_dir, f"{_m}{obj_suffix}") for _m in to_compile_lib])
        
        elif cxx == "cl" and platform.system() == "Windows":
            # TODO: to be implemented...
            ...

        else:
            raise NotImplementedError(f"Unsupported compiler: {cxx} in {platform.system()}")
        
    return module_name, script_dir, bin_dir

def _get_module_name(feat_dim):
    return f"vecdbImpl{feat_dim}"

def compile(
        feat_dim, 
        quite = False, 
        cxx: str = "g++",
        additional_compile_flags = [],
        additional_link_flags = []
        ) -> str:
    module_name, script_dir, bin_dir = _writeNinja(
        feat_dim, 
        cxx=cxx, 
        additional_compile_flags=additional_compile_flags, 
        additional_link_flags=additional_link_flags
        )

    def print_(*args, **kwargs):
        if not quite:
            print(*args, **kwargs)
    SP_STDOUT = subprocess.DEVNULL if quite else sys.stdout

    print_("\033[1;30m", end="\r")
    print_("----------------------------------------")

    subprocess.check_call(["ninja", "-t", "commands"], cwd = script_dir, stdout=SP_STDOUT)
    print_("----------------------------------------")

    subprocess.check_call("ninja", cwd = script_dir, stdout=SP_STDOUT)

    with open(os.path.join(script_dir, "compile_commands.json"), "w") as f:
        # ninja -t compdb > compile_commands.json
        subprocess.check_call(["ninja", "-t", "compdb"], cwd = script_dir, stdout=f, stderr=SP_STDOUT)

    print_("\033[0m", end="\r")
    return f"{bin_dir}.{module_name}".split(os.path.sep)[-1]


from __future__ import annotations
import os, subprocess, platform, time, random
from typing import TYPE_CHECKING
from .config import CACHE_DIR
if TYPE_CHECKING:
    from .wrap import CompileConfig

def checkCommandExists(cmd: str) -> bool:
    if platform.system() == "Windows":
        return subprocess.call(["where", cmd], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
    return subprocess.call(["which", cmd], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0


def initEigenSrc(eigen_src_path: str, eigen_version: str = "3.4.0"):
    if not os.path.exists(eigen_src_path):
        os.makedirs(eigen_src_path)

    __downloading_lock_file = os.path.join(CACHE_DIR, "downloading.lock")
    while os.path.exists(__downloading_lock_file):
        print(f"Find lock file [{__downloading_lock_file}], waiting...")
        time.sleep(random.random() * 1 + 1)     # wait for 1~2 seconds, avoid resource competition

    if os.listdir(eigen_src_path) == [] or \
        not os.path.exists(os.path.join(eigen_src_path, "Eigen", "src", "Core")):

        print("Downloading Eigen...")
        if not checkCommandExists("git"):
            raise RuntimeError("git not found.")
        
        open(__downloading_lock_file, "w").close()
        try:
            subprocess.check_call([
                "git", "clone", "--depth=1", f"--branch={eigen_version}",
                "https://gitlab.com/libeigen/eigen.git", eigen_src_path]
                )
        except: pass
        finally:
            os.remove(__downloading_lock_file)

def autoCompileConfig() -> CompileConfig:
    """
    Return automatically detected compile config for the current environment. 
    Will check existence of compilers and files.
    """
    # TODO: improve this...
    cxx = None
    for _cxx in ["g++", "clang++"]:
        if checkCommandExists(_cxx):
            cxx = _cxx
            break
    if cxx is None:
        raise RuntimeError("No C++ compiler found. Please install g++ or clang++.")
    
    additional_compile_flags = [
        "-march=native", "-mtune=native",
    ]

    additional_link_flags = []
    
    return {
        "cxx": cxx,
        "additional_compile_flags": additional_compile_flags,
        "additional_link_flags": additional_link_flags
    }


import sqlite3
from threading import Lock

def lockRequire(lock):
    def _func(func):
        def wrapper(*args, **kwargs):
            lock.acquire()
            try:
                return func(*args, **kwargs)
            finally:
                lock.release()
        return wrapper
    return _func

class SqliteIO:
    _lock = Lock()

    def __init__(self, fpath: str) -> None:
        self.conn = sqlite3.connect(fpath)
        self.cur = self.conn.cursor()
    
    @lockRequire(_lock)
    def touchTable(self, name: str) -> None:
        # create if not exists, save string
        self.cur.execute(f"CREATE TABLE IF NOT EXISTS {name} (id TEXT PRIMARY KEY, vector TEXT)")
    
    @lockRequire(_lock)
    def deleteTable(self, name: str) -> None:
        # delete table
        self.cur.execute(f"DROP TABLE {name}")

    def getTableNames(self) -> list[str]:
        # get all table names
        ret = self.cur.execute("SELECT name FROM sqlite_master WHERE type='table'")
        return [i[0] for i in ret]
    
    def getTableData(self, name: str) -> tuple[list[str], list[str]]:
        # get all data in table
        res = self.cur.execute(f"SELECT * FROM {name}")
        ret = ([], [])
        for i in res:
            ret[0].append(i[0])
            ret[1].append(i[1])
        return ret

    @lockRequire(_lock)
    def insetToTable(self, name: str, id: str, enc_vector: list[str]) -> None:
        # insert one row to table, make sure id not exists
        self.cur.execute(f"INSERT INTO {name} VALUES (?, ?)", (id, enc_vector))
    
    @lockRequire(_lock)
    def updateTable(self, name: str, id: str, enc_vector: list[str]) -> None:
        # update one row to table, make sure id exists
        self.cur.execute(f"UPDATE {name} SET vector = ? WHERE id = ?", (enc_vector, id))
    
    @lockRequire(_lock)
    def deleteFromTable(self, name: str, id: str) -> None:
        # delete one row from table, make sure id exists
        self.cur.execute(f"DELETE FROM {name} WHERE id = ?", (id,))
    
    @lockRequire(_lock)
    def commit(self):
        self.conn.commit()


from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    packages=find_packages(),
    include_package_data = True,
    long_description=long_description,
    long_description_content_type="text/markdown",
)
